#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include <time.h>
#include "utilities.h"
#include "Ring.h"

// -------------------------------------------------------------------------------
// Calculating an approximate solution corresponding to new resolution orders through
// spectral interpolation
// -------------------------------------------------------------------------------
void v1_To_v2(parameters par1, derivs_2D v1, parameters par2, derivs_2D v2)
{
	int	ipot, idom, i_1, j_1, i_2, j_2, indx;
	double p[NMAX], *v21;
	parameters par21=par1;

	par21.nst[0] = 0;
	for(idom=0; idom<NDOM; idom++){
		par21.ns[idom]    = par2.ns[idom]; 
		par21.nt[idom]    = par1.nt[idom]; 
		par21.nst[idom+1] = par21.nst[idom] + par21.ns[idom]*par21.nt[idom];
	}
	v21 = dvector(0, NPOT*par21.nst[NDOM]-1);	

	for(idom=0; idom<NDOM; idom++){
		int ns_1=par1.ns[idom], nt_1=par1.nt[idom],
			ns_2=par2.ns[idom], nt_2=par2.nt[idom], Ns_2=ns_2-1, Nt_2=nt_2-1;
		for(ipot=0; ipot < NPOT; ipot++) {
			for(j_1=0; j_1<nt_1; j_1++){
				for(i_1=0; i_1<ns_1; i_1++){
					indx   = Index(par1, idom, ipot, i_1, j_1);
					p[i_1] = v1.d0[indx];
				}
				chebft_Extremes(p,ns_1,0);
				for(i_2=0; i_2<ns_2; i_2++) {
					double a=Pih*i_2/Ns_2, sa=sin(a), s=sa*sa;
					indx      = Index(par21, idom, ipot, i_2, j_1);
					v21[indx] = chebev(0., 1., p, ns_1, s);
				}
			}
			for(i_2=0; i_2<ns_2; i_2++){
				for(j_1=0; j_1<nt_1; j_1++){
					indx   = Index(par21, idom, ipot, i_2, j_1);
					p[j_1] = v21[indx];
				}
				chebft_Extremes(p,nt_1,0);
				for(j_2=0; j_2<nt_2; j_2++) {
					double b=Pih*j_2/Nt_2, sb=sin(b), t=sb*sb;
					indx     = Index(par2, idom,ipot,i_2, j_2);
					v2.d0[indx] = chebev(0., 1., p, nt_1, t);
				}
			}
		}
	}	
	free_dvector(v21, 0, NPOT*par21.nst[NDOM]-1);
}
// -------------------------------------------------------------------------------
void wG1_To_wG2(parameters par1, derivs_1D w1, derivs_1D G1, parameters par2, derivs_1D w2, derivs_1D G2) 
{
	int	idom=NDOM-1, ns_1=par1.ns[idom], ns_2=par2.ns[idom], Ns_2=ns_2-1, ipot, i_1, i_2;
	double p[NMAX];
	
	for(ipot=0; ipot < NPOT; ipot++) {
		for(i_1=0; i_1<ns_1; i_1++){
			if (ipot==0) p[i_1] = G1.d0[i_1];
			else         p[i_1] = w1.d0[Index_Ring(par1, ipot, i_1)];
		}
		chebft_Extremes(p,ns_1,0);
		for(i_2=0; i_2<ns_2; i_2++) {
			double a=Pih*i_2/Ns_2, sa=sin(a), s=sa*sa,
				f=chebev(0., 1., p, ns_1, s);
			if (ipot==0) G2.d0[i_2]                         = f;
			else         w2.d0[Index_Ring(par2, ipot, i_2)] = f;
		}
	}
}
//   -----------------------------------------------
void GetNewResolution(parameters par1, parameters *par2, double **X1)
{
	int idom, ipot, ipar, i, j, ns, n, n_v;
	double *X2;
	derivs_2D v1, v2; 
	derivs_1D G1, w1, G2, w2; 

	ns = par1.ns[NDOM-1];
	allocate_derivs_1D(&G1, ns);
	allocate_derivs_1D(&w1, ns*NPOT);
	allocate_derivs_2D(&v1, par1.n_v_total);
	Get_vwGpar_From_X(par1, *X1, v1, w1, G1, &par1);
	
	X2 = dvector(0, (*par2).ntotal-1);
	ns = (*par2).ns[NDOM-1];
	allocate_derivs_1D(&G2, ns);
	allocate_derivs_1D(&w2, ns*NPOT);
	allocate_derivs_2D(&v2, (*par2).n_v_total);
	
	for(ipar=0; ipar<NPAR; ipar++)
		X2[(*par2).n_2D+ipar] = (*X1)[par1.n_2D+ipar];

	v1_To_v2(par1, v1, *par2, v2);
	wG1_To_wG2(par1, w1, G1, *par2, w2, G2); 
	
	for(n=0; n<(*par2).n_2D; n++){
		n_v  = (*par2).n_v_of_n[n];
		Get_Indices_From_n_v(*par2, n_v, &idom, &ipot, &i, &j, &ipar);
		
		if (idom < NDOM-1 || (idom == NDOM-1 && j > 0)) X2[n] = v2.d0[n_v];
		else{
			if(ipot > 0) X2[n] = w2.d0[Index_Ring(*par2, ipot, i)];
			else		 X2[n] = G2.d0[i];
		}
	}
	Get_vwGpar_From_X(*par2, X2, v2, w2, G2, par2);
	GetPhysicalParameters_NonLin(par2, v2, w2, G2);

	Free_Arrays_n_And_n_v(&par1);
	free_dvector(*X1, 0, par1.ntotal-1);

	ns = (*par2).ns[NDOM-1];
	free_derivs_2D(&v2, (*par2).n_v_total);
	free_derivs_1D(&w2, ns*NPOT);
	free_derivs_1D(&G2, ns);
	
	ns = par1.ns[NDOM-1];
	free_derivs_2D(&v1, (*par2).n_v_total);
	free_derivs_1D(&w1, ns*NPOT);
	free_derivs_1D(&G1, ns);
	
	*X1   = X2;
}
